@@ -16,6 +16,7 @@
 
 import datetime
 import logging
+import os
 import random
 import re
 import sys
@@ -24,13 +25,13 @@ from pathlib import Path
 from typing import Optional
 from typing import Tuple
 from xml.etree import ElementTree
+from xml.etree.ElementTree import ParseError
 
 import requests
 import requests_toolbelt
 import xmltodict
 from lxml import etree
 from lxml.etree import Element
-import os
 from pan import xapi
 from pan.config import PanConfig
 from pan.xapi import PanXapiError
@@ -64,6 +65,7 @@ class Panoply:
         """
         Initialize a new panoply object. Passing in the authentication information will cause this class to attempt
         to connect to the device and set offline_mode to False. Otherwise, offline mode will be set to True
+
         :param hostname: hostname or ip address of target device`
         :param api_username: username
         :param api_password: password
@@ -130,6 +132,7 @@ class Panoply:
     def connect(self, allow_offline: Optional[bool] = False) -> None:
         """
         Attempt to connect to this device instance
+
         :param allow_offline: Do not raise an exception if this device is offline unless there is an authentication
         error
         :return: None
@@ -166,8 +169,9 @@ class Panoply:
     def commit(self, force_sync=True) -> str:
         """
         Perform a commit operation on this device instance -
+
         :raises PanoplyException: if commit failed
-        :param force_sync: Flag to enablle synch commit or async
+        :param force_sync: Flag to enable sync commit or async
         :return: String from the API indicating success or failure
         """
 
@@ -175,21 +179,8 @@ class Panoply:
             self.xapi.commit(cmd='<commit></commit>', sync=force_sync, timeout=600)
             results = self.xapi.xml_result()
 
-            if results is None:
-                return self.xapi.status_detail
-
-            if force_sync:
-                doc = ElementTree.XML(results)
-                embedded_result = doc.find('result')
-
-                if embedded_result is not None:
-                    commit_result = embedded_result.text
-
-                    if commit_result == 'FAIL':
-                        raise PanoplyException(self.xapi.status_detail)
-
-                    else:
-                        return self.xapi.status_detail
+            if not self.__check_commit_return(results, force_sync):
+                raise PanoplyException(self.xapi.status_detail)
 
             return self.xapi.status_detail
 
@@ -197,9 +188,106 @@ class Panoply:
             logger.error(pxe)
             raise PanoplyException('Could not commit configuration')
 
+    def commit_gpcs(self, force_sync=True) -> str:
+        """
+        Perform a commit operation on this device instance specifically for gpcs remote networks
+        Note - you must do a full commit to panorama before you invoke this commit!
+
+        :raises PanoplyException: if commit failed
+        :param force_sync: Flag to enable sync commit or async
+        :return: String from the API indicating success or failure
+        """
+        try:
+
+            running_config = self.get_configuration(config_source='running')
+            config_doc = etree.fromstring(running_config)
+
+            sc = config_doc.xpath("/config/devices/entry[@name='localhost.localdomain']/"
+                                  "device-group/entry[@name='Service_Conn_Device_Group']")
+            rn = config_doc.xpath("/config/devices/entry[@name='localhost.localdomain']/"
+                                  "device-group/entry[@name='Remote_Network_Device_Group']")
+            mu = config_doc.xpath("/config/devices/entry[@name='localhost.localdomain']/"
+                                  "device-group/entry[@name='Mobile_User_Device_Groupp']")
+
+            if sc:
+                self.xapi.commit(action='all',
+                                 cmd='<commit-all><shared-policy><device-group>'
+                                     '<entry name="Service_Conn_Device_Group"/>'
+                                     '</device-group></shared-policy></commit-all>')
+
+                results = self.xapi.xml_result()
+
+                if not self.__check_commit_return(results, force_sync):
+                    raise PanoplyException(self.xapi.status_detail)
+
+            if rn:
+                self.xapi.commit(action='all',
+                                 cmd='<commit-all><shared-policy><device-group>'
+                                     '<entry name="Remote_Network_Device_Group"/>'
+                                     '</device-group></shared-policy></commit-all>')
+
+                results = self.xapi.xml_result()
+
+                if not self.__check_commit_return(results, force_sync):
+                    raise PanoplyException(self.xapi.status_detail)
+
+            if mu:
+                self.xapi.commit(action='all',
+                                 cmd='<commit-all><shared-policy><device-group>'
+                                     '<entry name="Mobile_User_Device_Group"/>'
+                                     '</device-group></shared-policy></commit-all>')
+
+                results = self.xapi.xml_result()
+
+                if not self.__check_commit_return(results, force_sync):
+                    raise PanoplyException(self.xapi.status_detail)
+
+            return 'Prisma Access Committed Successfully'
+
+        except PanXapiError as pxe:
+            logger.error(pxe)
+            raise PanoplyException('Could not commit configuration')
+
+    @staticmethod
+    def __check_commit_return(results: str, force_sync: bool) -> bool:
+        """
+        Check xml result from a panos device and check for a failure condition
+
+        :param results: str returned from xapi.xml_result
+        :param force_sync: if force_sync is true, then verify the commit actually succeeded
+        :return: boolean
+        """
+
+        if results is None:
+            return False
+
+        if force_sync:
+            embedded_result = None
+            try:
+                doc = ElementTree.XML(results)
+                embedded_result = doc.find('result')
+
+            except ParseError:
+                # results can be malformed xml from gpcs
+                if 'jobid' in results:
+                    return True
+
+            if embedded_result is not None:
+                commit_result = embedded_result.text
+
+                if commit_result == 'FAIL':
+                    return False
+
+                else:
+                    return True
+        else:
+            # fixme - is there something else we can check here?
+            return True
+
     def set_at_path(self, name: str, xpath: str, xml_str: str) -> None:
         """
         Insert XML into the configuration tree at the specified xpath
+
         :param name: name of the snippet - used in logging and debugging only
         :param xpath: full xpath where the xml element will be inserted
         :param xml_str: string representation of the XML element to insert
@@ -215,24 +303,46 @@ class Panoply:
         except PanXapiError as pxe:
             raise PanoplyException(f'Could not push skillet {name} / snippet {xpath}! {pxe}')
 
-    def execute_op(self, cmd_str: str, cmd_xml=False) -> str:
+    def execute_op(self, cmd_str: str, cmd_xml=False, parse_result=True) -> str:
         """
         Executes an 'op' command on the NGFW
+
         :param cmd_str: op command to send
         :param cmd_xml: Flag to determine if op command requires XML encoding
+        :param parse_result: Optional flag to indicate whether to return parsed xml results (xml_result) from xapi - not
+        all commands return valid XML. Setting this to 'false' will return the raw string from the API.
         :return: raw output from the device
         """
 
         try:
             self.xapi.op(cmd=cmd_str, cmd_xml=cmd_xml)
-            return self.xapi.xml_result()
+            if parse_result:
+                return self.xapi.xml_result()
+            else:
+                return self.xapi.xml_document
 
         except PanXapiError as pxe:
+            if 'ParseError' in str(pxe):
+                # some op commands do not properly wrap data in CDATA tags, so let's try to
+                # replace some chars manually and re-parse...
+                x = self.xapi.xml_document
+                cx = x.replace('&', '&amp;')
+
+                try:
+                    # make sure our little 'fix' didn't completely ruin the xml structure and return what we got
+                    s = ElementTree.fromstring(cx)
+                    return ElementTree.tostring(s)
+
+                except Exception as e:
+                    logger.error(e)
+                    raise PanoplyException(pxe)
+
             raise PanoplyException(pxe)
 
     def execute_cli(self, cmd_str: str) -> str:
         """
         Short-cut to execute a simple CLI op cmd
+
         :param cmd_str: CLI command to send to the NGFW
         :return: raw output from the command
         """
@@ -241,6 +351,7 @@ class Panoply:
     def execute_cmd(self, cmd: str, params: dict, context=None) -> str:
         """
         Execute the given cmd using the xapi.
+
         :param cmd: Valid options are: 'op', 'show', 'get', 'delete', 'set', 'edit', 'override', 'move', 'rename',
                                        'clone', 'validate'
         :param params: valid parameters for the given cmd type
@@ -252,13 +363,17 @@ class Panoply:
 
         # this code happily borrowed from ansible-pan module
         # https://raw.githubusercontent.com/PaloAltoNetworks/ansible-pan/develop/library/panos_type_cmd.py
-        cmd = params['cmd']
+
         func = getattr(self.xapi, cmd)
 
         # shortcut for op cmd types
         if cmd == 'op':
             cmd_str = ''.join(params['cmd_str'].strip().split('\n'))
-            return self.execute_op(cmd_str)
+
+            # fix for GL #79
+            parse_result = params.get('parse_result', True)
+
+            return self.execute_op(cmd_str, cmd_xml=False, parse_result=parse_result)
 
         # in all other cases, the xpath is a required attribute
         kwargs = {
@@ -301,6 +416,7 @@ class Panoply:
     def fetch_license(self, auth_code: str) -> bool:
         """
         Fetch and install licenses for PAN-OS NGFW
+
         :param auth_code: Authorization code to use to license the NGFW
         :return: True when license installation succeeds / False otherwise
         """
@@ -345,6 +461,7 @@ class Panoply:
     def set_license_api_key(self, api_key: str) -> bool:
         """
         Set's the Palo Alto Networks Support API Key in the firewall. This is required to deactivate a VM-Series NGFW.
+
         :param api_key: Your support account API Key found on the support.paloaltonetworks.com site
         :return: boolean True on success / False otherwise
         """
@@ -389,6 +506,7 @@ class Panoply:
     def deactivate_vm_license(self, api_key: str = None) -> bool:
         """
         Deactivate VM-Series Licenses. Will set the API Key is not already set.
+
         :param api_key: Optional api_key.
         :return: boolean True on success / False otherwise
         """
@@ -407,8 +525,8 @@ class Panoply:
                 return False
 
         try:
-            cmd = f'<request><license><deactivate><VM-Capacity><mode>auto</mode>' \
-                  f'</VM-Capacity></deactivate></license></request>'
+            cmd = '<request><license><deactivate><VM-Capacity><mode>auto</mode>' \
+                  '</VM-Capacity></deactivate></license></request>'
             logger.debug(f'Using request cmd: {cmd}')
             results = self.execute_op(cmd)
             logger.debug(f'deactivate_vm_license results: {results}')
@@ -433,6 +551,7 @@ class Panoply:
     def sanitize_element(element: str) -> str:
         """
         Eliminate some unneeded characters from the XML snippet if they appear.
+
         :param element: element str
         :return: sanitized element str
         """
@@ -444,6 +563,7 @@ class Panoply:
     def backup_config(self):
         """
         Saves a named backup on the PAN-OS device. The format for the backup is 'panhandler-20190424000000.xml'
+
         :return:  xml results from the op command sequence
         """
         d = datetime.datetime.today()
@@ -464,6 +584,7 @@ class Panoply:
         """
         Gather system info and keep on self.facts
         This gets called on every connect
+
         :return: dict containing all system facts
         """
         facts = {}
@@ -503,6 +624,7 @@ class Panoply:
         Load baseline config that contains ONLY connecting username / password
         use device facts to determine which baseline template to load
         see template/panos/baseline_80.xml for example
+
         :param self:
         :return: bool true on success
         """
@@ -517,6 +639,7 @@ class Panoply:
         Load baseline config that contains ONLY connecting username / password
         use device facts to determine which baseline template to load
         see template/panos/baseline_80.xml for example
+
         :param self:
         :param reset_hostname: Flag to reset hostname back to a default value (baseline in this case)
         :return: string contents of baseline config
@@ -577,7 +700,16 @@ class Panoply:
             skillet_dir = 'baseline_90'
 
         elif '9.1' in version:
-            # load the 9.0 baseline with
+            # load the 9.1 baseline with
+            skillet_dir = 'baseline_91'
+
+        # Support 9.2 Beta and 10.0 for GL #80
+        elif '9.2' in version:
+            # load the 9.1 baseline with
+            skillet_dir = 'baseline_91'
+
+        elif '10.0' in version:
+            # load the 9.1 baseline with
             skillet_dir = 'baseline_91'
 
         else:
@@ -599,6 +731,7 @@ class Panoply:
     def import_file(self, filename: str, file_contents: (str, bytes), category: str) -> bool:
         """
         Import the given file into this device
+
         :param filename:
         :param file_contents:
         :param category: 'configuration'
@@ -637,6 +770,7 @@ class Panoply:
     def load_config(self, filename: str) -> bool:
         """
         Loads the named configuration file into this device
+
         :param filename: name of the configuration file on the device to load. Note this filename must already exist
         on the target device
         :return: bool True on success
@@ -654,6 +788,11 @@ class Panoply:
     def wait_for_device_ready(self, interval=30, timeout=600) -> bool:
         """
         Loop and wait until device is ready or times out
+
+        .. note::
+
+            This currently only supports PAN-OS devices and not Panorama
+
         :param interval: how often to check in seconds
         :param timeout: how long to wait until we declare a timeout condition
         :return: boolean true on ready, false on timeout
@@ -685,9 +824,59 @@ class Panoply:
             logger.info(f'Waiting for {self.hostname} to become ready...')
             time.sleep(interval)
 
+    def filter_connected_devices(self, filter_terms=None) -> list:
+        """
+        Returns the list of connected devices filtered according to the given terms.
+
+        Filter terms are based on keys in the device facts. Matches are done using simple regex match
+
+        :param filter_terms: dict containing key value pairs. Keys match keys from the return device facts and values
+        are regex expressions used to match those keys
+        :return: list of devices that match ALL filter terms
+        """
+
+        if filter_terms is None:
+            filter_terms = {}
+
+        if self.facts.get('model', '') != 'Panorama':
+            logger.info('Method only supported on Panorama Devices')
+            return list()
+
+        connected_devices_xml = self.execute_cli('show devices connected')
+        connected_devices_dict = xmltodict.parse(connected_devices_xml)
+
+        if 'devices' in connected_devices_xml and 'entry' in connected_devices_dict['devices']:
+            connected_devices = connected_devices_dict['devices']['entry']
+
+        else:
+            return list()
+
+        if not filter_terms:
+            # no terms given, return all devices
+            return connected_devices
+
+        filtered_devices = list()
+
+        for device in connected_devices:
+            match = False
+
+            for term in filter_terms:
+                if term in device:
+
+                    if re.match(filter_terms[term], device[term]):
+                        match = True
+                    else:
+                        match = False
+
+            if match:
+                filtered_devices.append(device)
+
+        return filtered_devices
+
     def update_dynamic_content(self, content_type: str) -> bool:
         """
         Check for newer dynamic content and install if found
+
         :param content_type: type of content to check. can be either: 'content', 'anti-virus', 'wildfire'
         :return: bool True on success
         """
@@ -728,7 +917,7 @@ class Panoply:
                     raise PanoplyException('Could not install dynamic content')
 
             else:
-                logger.info(f'No job returned to track')
+                logger.info('No job returned to track')
 
             return True
 
@@ -740,6 +929,7 @@ class Panoply:
         """
         Iterate through all available content of the specified type, locate and return the version with the highest
         version number. If that version is already installed, return None as no further action is necessary
+
         :param content_type: type of content to check
         :return: version-number to download and install or None if already at the latest
         """
@@ -780,6 +970,7 @@ class Panoply:
         """
         Loops until a given job id is completed. Will timeout after the timeout period if the device is
         offline or otherwise unavailable.
+
         :param job_id: id the job to check and wait for
         :param interval: how long to wait between checks
         :param timeout: how long to wait with no response before we give up
@@ -833,6 +1024,7 @@ class Panoply:
     def get_configuration(self, config_source='running') -> str:
         """
         Get the configuration from the device.
+
         :return: configuration xml as a string or a blank string if not connected
         """
 
@@ -855,13 +1047,14 @@ class Panoply:
                 return ''
 
         except PanXapiError:
-            logger.error(f'Could not get configuration from device')
+            logger.error('Could not get configuration from device')
             raise PanoplyException('Could not get configuration from the device')
 
     def generate_skillet(self, from_candidate=False) -> list:
         """
         Generates a skillet from the changes detected on this device.
         This will attempt to create the xml and xpaths for everything that is found to have changed
+
         :param from_candidate: If your changes on in the candidate config, this will detect changes between the running
         config and the candidate config. If False, this will detect changes between the running config and a generic
         baseline configuration
@@ -882,15 +1075,24 @@ class Panoply:
         return self.generate_skillet_from_configs(previous_config, latest_config)
 
     def generate_skillet_from_configs(self, previous_config: str, latest_config: str) -> list:
+        """
+        Generates a skillet from the diffs between two XML configuration files
+
+        :param previous_config: Configuration backup taken before desired changes are made to the device
+        :param latest_config: Configuration backup taken after desired changes are made
+
+        :return: list of dictionaries that contain the following keys:
+            * name
+            * element
+            * xpath
+            * full_xpath
+        """
         # convert the config string to an xml doc
         latest_doc = etree.fromstring(latest_config)
 
         # let's grab the previous as well
         previous_doc = etree.fromstring(previous_config)
 
-        ignored_xpaths = [
-            '/config/mgt-config/users/entry[@name="admin"]'
-        ]
         current_xpath = '.'
         not_found_xpaths = list()
 
@@ -909,8 +1111,8 @@ class Panoply:
             full_relative_xpath = re.sub(r'^/config/', r'\./', xpath)
             tag = re.sub(r'\[.*\]', '', entry)
 
-            if set_xpath in ignored_xpaths:
-                logger.debug(f'Skipping ignored xpath: {xpath}')
+            # check if this xpath is actually user configurable
+            if self.__is_ignored_xpath(set_xpath):
                 continue
 
             changed_elements = latest_doc.xpath(xpath)
@@ -934,12 +1136,34 @@ class Panoply:
 
         return self.__order_snippets(snippets)
 
+    @staticmethod
+    def __is_ignored_xpath(set_xpath: str) -> bool:
+        """
+        Determines if the given set_xpath should be used for skillet generation purposes. Certain xpaths will be
+        changed, however, they are not user configurable, and therefore should not be included
+
+        :param set_xpath: xpath string to check
+        :return: boolean true if the xpath should be skipped
+        """
+        ignored_xpaths = (
+            '/config/mgt-config/users/entry[@name="admin"]',
+            '/config/shared/content-preview'
+        )
+
+        for ignored_xpath in ignored_xpaths:
+            if ignored_xpath in set_xpath:
+                logger.debug(f'Skipping ignored xpath: {set_xpath}')
+                return True
+
+        return False
+
     @staticmethod
     def generate_set_cli_from_configs(previous_config: str, latest_config: str) -> list:
         """
         Takes two configuration files, converts them to set commands, then returns only the commands found
         in the 'latest_config' vs the 'previous_config'. This allows the user to quickly configure one firewall,
         generate a 'set cli' diff and move those configs to another firewall
+
         :param previous_config: Starting config
         :param latest_config: Ending config
         :return: list of set cli commands required to convert previous to latest
@@ -968,6 +1192,7 @@ class Panoply:
         """
         recursive function to determine if the 'el' Element found at 'xpath' can also be found at the same
         xpath in the 'pc' previous_config. Keep tabs on what has not been found using the 'not_founds' list
+
         :param el: The element in question from the latest_config
         :param xpath: the xpath to the element in question
         :param pc:  the previous config Element
@@ -1022,7 +1247,8 @@ class Panoply:
                         if k != 'uuid':
                             attribs.append(f'@{k}="{v}"')
 
-                    attrib_str = " ".join(attribs)
+                    # fix for #71
+                    attrib_str = " and ".join(attribs)
                     # track the attributes in the xpath by virtue of the 'path_entry' which will be appended to the
                     # xpath later
                     path_entry = f'{e.tag}[{attrib_str}]'
@@ -1080,12 +1306,24 @@ class Panoply:
             '/shared/',  # catch the rest of the shared items here
             '/tag/entry',
             '/deviceconfig/system',
-            '/network/interface', '/network/virtual-wire', '/network/vlan',
-            '/network/ike', '/network/tunnel',
-            '/network/virtual-router', '/network/profiles/zone-protection-profile', '/zone/entry',
+            '/network/profiles',
+            '/network/interface',
+            '/network/virtual-wire',
+            '/network/vlan',
+            '/network/ike',
+            '/network/tunnel',
+            '/import/network/interface'  # for use with panorama plugin import (SD-WAN)
+            '/network/virtual-router',
+            '/network/profiles/zone-protection-profile',
+            'routing-table/ip/static-route/entry/next-hop',  # try to keep next-hop before path-monitor for set cli
+            'routing-table/ip/static-route/entry/path-monitor',  # #70
+            '/dynamic-ip-and-port/interface-address',  # GH #70
+            '/dynamic-ip-and-port/interface',  # GH #70
+            '/dynamic-ip-and-port/ip',  # GH #70
+            '/zone/entry',
             '/profiles/custom-url-category',  # should come before profiles/url-filtering
             '/address/entry'  # should come before rules or address-group
-            ]
+        ]
 
         ordered_snippets = list()
         for x in xpaths:
@@ -1111,6 +1349,7 @@ class Panoply:
     def __filter_snippets_by_xpath(snippets: list, xpath: str) -> list:
         """
         Check the each snippet in the list and return those that has a full_xpath matching the xpath param
+
         :param snippets: list of snippets
         :param xpath: xpath to check
         :return: list of only those that contain the xpath in their full_xpath attribute
@@ -1167,9 +1406,6 @@ class Panoply:
         # convert the config string to an xml doc
         latest_doc = ElementTree.fromstring(latest_config)
 
-        # let's grab the previous as well
-        previous_doc = ElementTree.fromstring(previous_config)
-
         for d in diffs:
             logger.debug(d)
             # step 1 - find all inserted nodes (future enhancement can consider other types of detected changes as well
@@ -1415,6 +1651,10 @@ class Panoply:
 
 
 class EphemeralPanos(Panoply):
+    """
+    EphemeralPanos is used when a Panos instance may or may not be online and available at all times. This is useful
+    when instantiating instances via a CI/CD pipeline or other automation tool.
+    """
     pass
 
 
@@ -1426,13 +1666,12 @@ class Panos(Panoply):
 
     def __init__(self, hostname: Optional[str], api_username: Optional[str], api_password: Optional[str],
                  api_port: Optional[int] = 443, serial_number: Optional[str] = None,
-                 debug: Optional[bool] = False):
+                 debug: Optional[bool] = False, api_key: Optional[str] = None):
 
-        super().__init__(hostname, api_username, api_password, api_port, serial_number, debug)
+        super().__init__(hostname, api_username, api_password, api_port, serial_number, debug, api_key)
 
         if self.connected:
             return
 
         if self.xapi and 'URLError' in self.xapi.status_detail:
             raise TargetConnectionException(self.xapi.status_detail)
-
