@@ -16,6 +16,7 @@
 
 import json
 import logging
+import re
 import xml.etree.ElementTree as elementTree
 from abc import ABC
 from abc import abstractmethod
@@ -26,12 +27,15 @@ from xml.etree.ElementTree import ParseError
 import xmltodict
 from jinja2 import BaseLoader
 from jinja2 import Environment
+from jinja2 import meta
 from jinja2.exceptions import TemplateAssertionError
 from jinja2.exceptions import UndefinedError
 from jsonpath_ng import parse
 from lxml import etree
 from passlib.hash import md5_crypt
 
+from jinja2_ansible_filters import AnsibleCoreFiltersExtension
+
 from skilletlib.exceptions import SkilletLoaderException
 from skilletlib.exceptions import SkilletValidationException
 
@@ -40,16 +44,19 @@ logger = logging.getLogger(__name__)
 
 class Snippet(ABC):
     """
-    BaseSnippet implements a basic template object snippet
+    BaseSnippet implements a basic Noop snippet
     """
     # set of required metadata, each snippet will define what attributes are required in the snippet definition
     # by default, we only require a 'name' attribute, but sub-classes will require more
     required_metadata = {'name'}
 
     # dict of optional metadata  and their default values. These values will be set on the snippet class but
-    # will not throw an exeption is they are not present
+    # will not throw an exception is they are not present
     optional_metadata = dict()
 
+    # metadata fields that should be considered templates and rendered
+    template_metadata = set()
+
     # set a default output type. this can be overridden for each SnippetType. This is used to determine the default
     # output handler to use for each snippet class. This can be set on a per snippet basis, but this allows a
     # short-cut on each
@@ -82,37 +89,48 @@ class Snippet(ABC):
 
         self.context = dict()
 
-    def update_context(self, context: dict) -> None:
+    def update_context(self, context: dict) -> dict:
         """
         This will update the snippet context with the passed in dict.
-        This gets called before render_metadata
-        :param context:
-        :return:
+        This gets called inside of 'should_execute'
+
+        :param context: dict of the outer context
+        :return: newly updated context
         """
         self.context.update(context)
+        return self.context
 
     @abstractmethod
-    def execute(self, context: dict) -> Tuple[dict, str]:
+    def execute(self, context: dict) -> Tuple[str, str]:
         """
-        Execute this Snippet and return a tuple consisting on the updated context and a string representing
+        Execute this Snippet and return a tuple consisting on raw output and a string representing
         success, failure, or running.
 
         Each snippet sub class must override this method!
 
         :param context: context to use for variable interpolation
-        :return: Tuple containng updated context dictionary and string indicated success or failure
+        :return: Tuple containing raw snippet output and string indicated success or failure
         """
-        return dict(), 'success'
+        return '', 'success'
 
     def should_execute(self, context: dict) -> bool:
         """
         Evaluate 'when' conditionals and return a bool if this snippet should be executed
+
         :param context: jinja context containing previous outputs and user supplied variables
         :return: boolean
         """
 
         logger.debug(f'Checking snippet: {self.name}')
 
+        # hook for pre-conditional checks
+        context = self.update_context(context)
+
+        # check if this snippet should be filtered out of execution based on the context object __filter_snippets
+        if self.is_filtered(context):
+            logger.debug('Skipping this snippet due to snippet inclusion rules')
+            return False
+
         if 'when' not in self.metadata:
             # always execute when no when conditional is present
             logger.debug(f'No conditional present, proceeding with skillet: {self.name}')
@@ -122,9 +140,113 @@ class Snippet(ABC):
         logger.debug(f'  Conditional Evaluation results: {results} ')
         return results
 
+    def is_filtered(self, context) -> bool:
+        """
+        Determines if a snippet should be available for execution based on the presence of the `__filter_snippets`
+        object in the context. Snippets can be filtered by the following:
+
+        include_by_name: list of names to check. Only snippet names included in this list will be executed
+        include_by_tag: list of tags to check. Only snippets with those tags will be executed
+        include_by_regex: regular expression match. Only snippets whose name matches the regex will be executed
+
+        any snippet that does not match any of the above rules will be filtered out. The rules are inclusive OR
+
+        :param context: Snippet Context
+        :return: bool
+        """
+
+        # by default, nothing is filtered out
+        is_filtered = None
+
+        if '__filter_snippets' in context:
+            logger.debug('filtering snippet...')
+
+            fc = context['__filter_snippets']
+            if type(fc) is not dict:
+                logger.warning('Snippet filter is malformed...')
+                return False
+            for filter_def in ('include_by_tag', 'include_by_name', 'include_by_regex', 'exclude_by_tag',
+                               'exclude_by_name', 'exclude_by_regex'):
+                if filter_def in fc:
+                    if type(fc[filter_def]) is list:
+                        for item in fc[filter_def]:
+                            is_filtered = self.__consider_filter(filter_def, item)
+                    elif type(fc[filter_def]) is str:
+                        item = fc[filter_def]
+                        is_filtered = self.__consider_filter(filter_def, item)
+
+                # we have discovered this snippet should not be filtered, jump out now
+                if is_filtered is not None:
+                    return is_filtered
+
+            # we have considered all rules, and still no determination, however, because we have some rules due to
+            # presence of '__filter_snippets' any item that is not specifically included or excluded, should be
+            # excluded
+
+            if is_filtered is None:
+                return True
+        else:
+            # there is not filter snippets config present, so include all snippets by default
+            return False
+
+    def __consider_filter(self, filter_def: str, item: str) -> (bool, None):
+        """
+        Rules that positively match are returned immediately, any snippet that does NOT specifically match any rule
+        is excluded (True response == snippet is NOT executed)
+
+        :param filter_def: type of filter to consider
+        :param item: specific filter item
+        :return: bool or None if not match
+        """
+        if filter_def == 'include_by_name':
+            # this name matches, do NOT filter it out
+            if self.name == item:
+                return False
+
+        elif filter_def == 'exclude_by_name':
+            # name matches exclusion rule, filter it out
+            if self.name == item:
+                return True
+
+        if filter_def == 'include_by_tag':
+            if self.__has_tag(item):
+                # snippet has this tag, do not filter out
+                return False
+
+        elif filter_def == 'exclude_by_tag':
+            if self.__has_tag(item):
+                # snippet has this tag, filter it out
+                return True
+
+        if '_by_regex' in filter_def:
+            match = re.match(item, self.name)
+            if 'include_by_regex' == filter_def:
+                if match:
+                    return False
+            elif 'exclude_by_regex' == filter_def:
+                if match:
+                    return True
+
+        # this snippet does not match any of the rules above, one way or the other, so filter it out of consideration
+        return None
+
+    def __has_tag(self, tag_to_check: str):
+        if 'tag' in self.metadata and type(self.metadata['tag']) is list:
+            for tag in self.metadata['tag']:
+                if tag_to_check == tag:
+                    return True
+
+            return False
+
+        elif 'tag' in self.metadata and type(self.metadata['tag']) is str:
+            return self.metadata['tag'] == tag_to_check
+        else:
+            return False
+
     def execute_conditional(self, test: str, context: dict) -> bool:
         """
         Evaluate 'test' conditionals and return a bool
+
         :param test: string of the conditional to execute
         :param context: jinja context containing previous outputs and user supplied variables
         :return: boolean
@@ -157,7 +279,8 @@ class Snippet(ABC):
         """
         get_output can be used when a snippet executes async and cannot or will not return output right away
         snippets that operate async must override this method
-        :return:
+
+        :return: Tuple containing the skillet output as a str and a str indicating success of failure
         """
 
         return '', 'success'
@@ -166,6 +289,7 @@ class Snippet(ABC):
         """
         each snippet type can override this method to provide it's own default output. This is used
         when there are no variables defined to be captured
+
         :param results: raw output from snippet execution
         :param status: status of the snippet.execute method
         :return: dict of default outputs
@@ -182,6 +306,7 @@ class Snippet(ABC):
     def capture_outputs(self, results: str, status: str) -> dict:
         """
         All snippet output or portions of snippet output can be captured and saved on the context as a new variable
+
         :param results: the raw output from the snippet execution
         :param status: status of the snippet.execute method
         :return: a dictionary containing all captured variables
@@ -208,6 +333,10 @@ class Snippet(ABC):
             if 'name' not in output:
                 continue
 
+            if not results:
+                outputs[output['name']] = ''
+                continue
+
             if 'capture_variable' in output:
                 outputs[output['name']] = self.render(output['capture_variable'], self.context)
             else:
@@ -240,7 +369,8 @@ class Snippet(ABC):
         return captured_outputs
 
     def __render_output_metadata(self, output: dict, context: dict) -> dict:
-        keys = ('capture_value', 'capture_pattern', 'capture_object', 'capture_list')
+        # fix for #78 allow filter_items to be rendered
+        keys = ('capture_value', 'capture_pattern', 'capture_object', 'capture_list', 'filter_items')
         for k in keys:
             if k in output:
                 output[k] = self.render(output[k], context)
@@ -250,6 +380,7 @@ class Snippet(ABC):
     def __filter_outputs(self, output_definition: dict, output: (str, dict, list), local_context: dict) -> (list, None):
         """
         Filter OUT items that do not pass the test
+
         :param output_definition: the output definition from the skillet
         :param output: the captured object to test
         :param local_context: local context for the jinja expression based tests
@@ -288,36 +419,106 @@ class Snippet(ABC):
         return filtered_items
 
     def render(self, template_str: str, context: (dict, None)) -> str:
+        """
+        Convenience method to quickly render a template_str using the provided context
+
+        :param template_str: jinja2 template to render
+        :param context: context to pass to the jinja2 environment
+        :return: rendered string
+        """
         if context is None:
             context = self.context
         t = self._env.from_string(template_str)
         return t.render(context)
 
+    def get_variables_from_template(self, template_str: str) -> list:
+        """
+        Returns a list of jinja2 variable found in the template
+
+        :param template_str: jinja2 template
+        :return: list of variables declared in the template
+        """
+
+        parsed_template_str = self._env.parse(template_str)
+        return meta.find_undeclared_variables(parsed_template_str)
+
+    def get_snippet_variables(self) -> list:
+        """
+        Returns a list of variables defined in this snippet
+
+        :return: list of str representing variable found in the jinja templates
+        """
+
+        variables = list()
+        for i in self.template_metadata:
+            if i in self.metadata:
+                variables.extend(self.get_variables_from_template(self.metadata[i]))
+
+        return variables
+
     def sanitize_metadata(self, metadata: dict) -> dict:
         """
-        Ensure the configured metadata is valid for this snippet type
-        :param metadata: dict
-        :return: validated metadata dict
-        """
-        name = metadata.get('name', '')
-        if not self.required_metadata.issubset(metadata):
-            for attr_name in metadata:
-                if attr_name not in self.required_metadata:
-                    raise SkilletLoaderException(f'Invalid snippet metadata configuration: attribute: {attr_name} '
-                                                 f'is required for snippet: {name}')
+        method to sanitize metadata. Each snippet type can override this provide extra logic over and above
+        just checking the required and optional fields
 
+        :param metadata: snippet metadata
+        :return: sanitized snippet metadata
+        """
         return metadata
 
     def render_metadata(self, context: dict) -> dict:
         """
         Each snippet sub class can override this method to perform jinja variable interpolation on various items
         in it's snippet definition. For example, the PanosSnippet will check the 'xpath' attribute and perform
-        the required interpolation
+        the required interpolation.
+
+        This handles regular strings, lists, and dictionaries such as:
+
+        in snippet class
+        template_metadata = {'render_me', 'render_all', 'render_list'}
+
+        in metadata
+        snippets:
+         - name: render_me_snippet
+           render_me: render_{{ this }}
+         - name: render_all_snippet
+           render_all:
+             a_key: some_{{ value }}
+             another_key: some_other_{{ value }}
+         - name: render_list_snippet
+           render_list:
+             - here_is_a_{{ value }}
+             - another_{{ value }}
+
         :param context: context from environment
         :return: metadata with jinja rendered variables
         """
         self.context.update(context)
 
+        # render all template metadata fields
+        for key_name in self.template_metadata:
+
+            if key_name in self.metadata:
+                key = self.metadata[key_name]
+
+                if isinstance(key, str):
+                    rendered_str = self.render(key, context)
+                    self.metadata[key_name] = rendered_str
+
+                elif isinstance(key, dict):
+                    rendered_dict = dict()
+                    for k, v in key.items():
+                        rendered_dict[k] = self.render(v, context)
+
+                    self.metadata[key_name] = rendered_dict
+
+                elif isinstance(key, list):
+                    rendered_list = list()
+                    for v in key:
+                        rendered_list.append(self.render(v, context))
+
+                    self.metadata[key_name] = rendered_list
+
         return self.metadata
 
     # define functions for custom jinja filters
@@ -325,6 +526,7 @@ class Snippet(ABC):
     def __md5_hash(txt: str) -> str:
         """
         Returns the MD5 Hashed secret for use as a password hash in the PAN-OS configuration
+
         :param txt: text to be hashed
         :return: password hash of the string with salt and configuration information. Suitable to place in the phash field
         in the configurations
@@ -335,16 +537,18 @@ class Snippet(ABC):
     def __init_env(self) -> None:
         """
         init the jinja2 environment and add any required filters
+
         :return: Jinja2 environment object
         """
-        self._env = Environment(loader=BaseLoader)
+        self._env = Environment(loader=BaseLoader, extensions=[AnsibleCoreFiltersExtension])
         self._env.filters["md5_hash"] = self.__md5_hash
         self.add_filters()
 
     def add_filters(self) -> None:
         """
         Each snippet sub-class can add additional filters. See the PanosSnippet for examples
-        :return:
+
+        :return: None
         """
         pass
 
@@ -363,7 +567,40 @@ class Snippet(ABC):
         """
         outputs = dict()
         output_name = output_definition.get('name', self.name)
-        outputs[output_name] = results
+        outputs[output_name] = ''
+
+        # enhancement for https://gitlab.com/panw-gse/as/skilletlib/-/issues/86
+        if 'capture_value' in output_definition:
+            # allow capture_value to be equivalent to capture_pattern
+            output_definition['capture_pattern'] = output_definition['capture_value']
+
+        if 'capture_object' in output_definition:
+            # allow capture_object to be equivalent to capture_list
+            output_definition['capture_list'] = output_definition['capture_object']
+
+        if 'capture_pattern' in output_definition:
+            # this is a regex pattern we should use for a match
+            pattern = re.compile(output_definition['capture_pattern'])
+            matches = pattern.findall(results)
+            if matches:
+                # capture pattern should only return the first match
+                outputs[output_name] = matches[0]
+
+        elif 'capture_list' in output_definition:
+            # this is a regex pattern we should use for a match
+            pattern = re.compile(output_definition['capture_list'])
+            matches = pattern.findall(results)
+            if matches:
+                # capture list should only the full list of matches
+                outputs[output_name] = matches
+            else:
+                # no matches should return an empty list
+                outputs[output_name] = list()
+
+        else:
+            output_name = output_definition.get('name', self.name)
+            outputs[output_name] = results
+
         return outputs
 
     def __handle_xml_outputs(self, output_definition: dict, results: str) -> dict:
@@ -383,6 +620,7 @@ class Snippet(ABC):
               - name: sw_version
                 capture_value: result/system/sw-version
 
+
         :param results: string as returned from some action, to be parsed as XML document
         :return: dict containing all outputs found from the capture pattern in each output
         """
@@ -432,7 +670,7 @@ class Snippet(ABC):
                 elif len(entries) == 1:
                     entry = entries.pop()
                     if isinstance(entry, str):
-                        captured_output[var_name] = entry
+                        captured_output[var_name] = str(entry)
                     else:
                         if len(entry) == 0:
                             # this tag has no children, so try to grab the text
@@ -516,10 +754,11 @@ class Snippet(ABC):
 
     def _handle_base64_outputs(self, results: str) -> dict:
         """
-         Parses results and returns a dict containing base64 encoded values
-         :param results: string as returned from some action, to be encoded as base64
-         :return: dict containing all outputs found from the capture pattern in each output
-         """
+        Parses results and returns a dict containing base64 encoded values
+
+        :param results: string as returned from some action, to be encoded as base64
+        :return: dict containing all outputs found from the capture pattern in each output
+        """
 
         outputs = dict()
 
@@ -548,8 +787,7 @@ class Snippet(ABC):
 
     def __handle_json_outputs(self, output_definition: dict, results: str) -> dict:
         """
-        Parses results and returns a dict containing base64 encoded values
-
+        Parses results using jsonpath_ng query syntax
         output_type: json
         outputs:
           - name: salt_auth_token
@@ -614,6 +852,7 @@ class Snippet(ABC):
     def __handle_manual_outputs(self, output_definition: dict, results: str) -> dict:
         """
         Manually set a value in the context, this could be useful with 'when' conditionals
+
         :param results: results from snippet execution, ignored in this method
         :return: dict containing manually defined name / value pair
         """
@@ -632,6 +871,7 @@ class Snippet(ABC):
     def _handle_manual_outputs(self, results: str) -> dict:
         """
         Manually set a value in the context, this could be useful with 'when' conditionals
+
         :param results: results from snippet execution, ignored in this method
         :return: dict containing manually defined name / value pair
         """
