@@ -34,13 +34,11 @@ logger = logging.getLogger(__name__)
 
 
 class PanosSnippet(TemplateSnippet):
+    # fields that are required in the metadata definition
     required_metadata = {'name'}
 
-    # optional metadata that may be overridden in the snippet definition / metadata
-    optional_metadata = {
-        'fail_message': 'Snippet Validation Failed',
-        'pass_message': 'Snippet Validation Passed'
-    }
+    # attribute fields that should be rendered during render_metadata
+    template_metadata = {'xpath', 'element', 'cmd_str'}
 
     # default output_type for each snippet of this type
     output_type = 'xml'
@@ -140,6 +138,7 @@ class PanosSnippet(TemplateSnippet):
     def sanitize_metadata(self, metadata: dict) -> dict:
         """
         Ensure all required keys are present in the snippet definition
+
         :param metadata: dict
         :return: dict
         """
@@ -200,6 +199,7 @@ class PanosSnippet(TemplateSnippet):
         """
         Renders each item in the metadata using the provided context.
         Currently renders the xpath and element for PANOS type skillets
+
         :param context: dict containing key value pairs to
         :return: dict containing the snippet definition metadata with the attribute values rendered accordingly
         """
@@ -214,15 +214,6 @@ class PanosSnippet(TemplateSnippet):
                                                            self.metadata['cherry_pick'])
                 meta['xpath'] = self.cherry_pick_xpath(self.metadata['xpath'],
                                                        self.metadata['cherry_pick'])
-            else:
-                if 'xpath' in self.metadata:
-                    meta['xpath'] = self.render(self.metadata['xpath'], context)
-
-                if 'element' in self.metadata:
-                    meta['element'] = self.render(self.metadata['element'], context)
-
-            if 'cmd_str' in self.metadata:
-                meta['cmd_str'] = self.render(self.metadata['cmd_str'], context)
 
         except TypeError as te:
             logger.info(f'Could not render metadata for snippet: {self.name}: {te}')
@@ -233,7 +224,9 @@ class PanosSnippet(TemplateSnippet):
     def compare_element_at_xpath(config: str, element: str, xpath: str, context: dict) -> bool:
         """
         Grab an xml fragment from the config given at xpath and compare it to this element
+
         :param config: XML document string from which to pull the XML element to compare
+
         :param element: element to check against
         :param xpath: xpath to grab an xml fragment from the config for comparison
         :param context: jinja context used to interpolate any variables that may be present in the template
@@ -262,10 +255,10 @@ class PanosSnippet(TemplateSnippet):
         """
         Cherry picking allows the skillet builder to pull out specific bits of a larger configuration
         and load only the smaller chunks. This is especially useful when combined with 'when' conditionals
+
         :param element: string containing the jinja templated xml fragment
         :param cherry_pick_path: string describing the relative xpath to use to cherry pick an xml node from the
             element given as a parameter
-        :param context: jinja context used to interpolate any variables that may be present in the template
         :return: rendered and cherry_picked element
         """
 
@@ -281,7 +274,7 @@ class PanosSnippet(TemplateSnippet):
             new_element = elementTree.tostring(cherry_picked_element).strip()
             return new_element
 
-        except ParseError as pe:
+        except ParseError:
             raise SkilletLoaderException(f'Could not parse element for cherry picking for snippet: {self.name}')
 
     def cherry_pick_xpath(self, base_xpath: str, cherry_picked_xpath: str) -> str:
@@ -356,6 +349,7 @@ class PanosSnippet(TemplateSnippet):
     def __node_attribute_present(self, obj: dict, config_path: str, attribute_name: str, attribute_value: str) -> bool:
         """
         Ensure a node with the named attribute and value does exist
+
         :param obj: obj to check
         :param config_path: path or child attribute(s) to check
         :param attribute_name: name of the attribute
@@ -387,6 +381,7 @@ class PanosSnippet(TemplateSnippet):
     def __node_attribute_absent(self, obj: dict, config_path: str, attribute_name: str, attribute_value: str) -> bool:
         """
         Ensure a node with the named attribute and value does not exist
+
         :param obj: obj to check
         :param config_path: path or child attribute(s) to check
         :param attribute_name: name of the attribute
@@ -473,6 +468,7 @@ class PanosSnippet(TemplateSnippet):
     def __append_uuid(self, string_input: str) -> str:
         """
         Simple filter to append a generated UUID to the end of the given string
+
         :param string_input: string to which to append the UUID
         :return: string with a dash and a string UUID
         """
@@ -487,6 +483,7 @@ class PanosSnippet(TemplateSnippet):
         and we need to know if the child key name exists on either the outer dict or the inner dict
         return the inner dict if the child key name is found there, or punt and return the outer dict
         otherwise
+
         :param obj: dict to check for child key name
         :param child: name of a key we want to find
         :return: inner dict if it contains the child key, outer dict otherwise
@@ -501,31 +498,19 @@ class PanosSnippet(TemplateSnippet):
 
         return obj
 
-    def handle_output_type_validation(self, results: str):
-
-        output = dict()
-        output['results'] = results
-        output['label'] = self.metadata.get('label', '')
-        output['severity'] = self.metadata.get('severity', 'low')
-        output['documentation_link'] = self.metadata.get('documentation_link', '')
-        output['test'] = self.metadata.get('test', '')
-
-        o = dict()
-        o[self.name] = output
-        return o
-
     def get_default_output(self, results: str, status: str) -> dict:
         """
         Override the default snippet get_default_output to not include raw results
+
         :param results: raw output from snippet execution
         :param status: status of the snippet.execute method
         :return: dict of default outputs
         """
 
         r = {
-                self.name: {
-                    'results': status,
-                    'changed': self.destructive
-                }
+            self.name: {
+                'results': status,
+                'changed': self.destructive
+            }
         }
         return r
